"""Evaluate polynomial."""
import numpy
import numpoly


def call(poly, *args, **kwargs):
    """
    Evaluate polynomial by inserting new values in to the indeterminants.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to evaluate.
        args (int, float, numpy.ndarray, numpoly.ndpoly):
            Argument to evaluate indeterminants. Ordered positional by
            ``poly.indeterminants``.
        kwargs (int, float, numpy.ndarray, numpoly.ndpoly):
            Same as ``args``, but positioned by name.

    Returns:
        (numpoly.ndpoly):
            Evaluated polynomial.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> poly = numpoly.polynomial([[x, x-1], [y, y+x]])
        >>> print(poly)
        [[x -1+x]
         [y y+x]]
        >>> print(poly(1, 0))
        [[1 0]
         [0 1]]
        >>> print(poly(1, y=[0, 1, 2]))
        [[[1 1 1]
          [0 0 0]]
        <BLANKLINE>
         [[0 1 2]
          [1 2 3]]]
        >>> print(poly(y))
        [[y -1+y]
         [y 2*y]]
        >>> print(poly(y=x-1, x=2*y))
        [[2*y -1+2*y]
         [-1+x -1+2*y+x]]

    """
    # Make sure kwargs contains all args and nothing but indeterminants:
    for arg, indeterminant in zip(args, poly.names):
        if indeterminant in kwargs:
            raise TypeError(
                "multiple values for argument '%s'" % indeterminant)
        kwargs[indeterminant] = arg
    extra_args = [key for key in kwargs if key not in poly.names]
    if extra_args:
        raise TypeError("unexpected keyword argument '%s'" % extra_args[0])

    if not kwargs:
        return poly.copy()

    # Saturate kwargs with values not given:
    indeterminants = poly.indeterminants
    for indeterminant in indeterminants:
        name = indeterminant.names[0]
        if name not in kwargs:
            kwargs[name] = indeterminant

    # There can only be one shape:
    ones = numpy.ones((), dtype=int)
    for value in kwargs.values():
        ones = ones * numpy.ones(numpoly.polynomial(value).shape, dtype=int)

    # main loop:
    out = 0
    for exponent, coefficient in zip(poly.exponents, poly.coefficients):
        term = ones
        for power, name in zip(exponent, poly.names):
            term = term*kwargs[name]**power
        shape = coefficient.shape+ones.shape
        out = out+numpoly.outer(coefficient, term).reshape(shape)

    if out.isconstant():
        return out.toarray()
    out, _ = numpoly.align_indeterminants(out, indeterminants)
    return out



# pylint: disable=wildcard-import
"""Numpoly -- Multivariate polynomials as numpy elements."""
from .baseclass import ndpoly
from .exponent import exponents_to_keys, keys_to_exponents

from .align import (
    align_polynomials,
    align_exponents,
    align_indeterminants,
    align_shape,
    align_dtype,
)
from .construct import (
    polynomial,
    aspolynomial,
    clean_attributes,
)
from .sympy_ import to_sympy

from .array_function import *
from .poly_function import *



"""Convert the input to an polynomial array."""
import numpoly


def aspolynomial(
        poly_like=None,
        indeterminants=None,
        dtype=None,
):
    """
    Convert the input to an polynomial array.

    Args:
        poly_like (Any):
            Input to be converted to a `numpoly.ndpoly` polynomial type.
        indeterminants (str, Tuple[str, ...]):
            Name of the indeterminant variables. If possible to infer from
            ``poly_like``, this argument will be ignored.
        dtype (type, numpy.dtype):
            Data type used for the polynomial coefficients.

    Returns:
        (numpoly.ndpoly):
            Array interpretation of `poly_like`. No copy is performed if the
            input is already an ndpoly with matching indeterminants and dtype.

    Examples:
        >>> x = numpoly.symbols("x")
        >>> numpoly.polynomial(x) is x
        False
        >>> numpoly.aspolynomial(x) is x
        True

    """
    remain = False
    if isinstance(poly_like, numpoly.ndpoly):

        remain = (dtype is None or dtype == poly_like.dtype)
        if indeterminants is not None:
            if isinstance(indeterminants, numpoly.ndpoly):
                indeterminants = indeterminants.names
            if isinstance(indeterminants, str):
                indeterminants = [indeterminants]
            if len(indeterminants) == 1 and len(poly_like.names) > 1:
                indeterminants = [
                    "{}{}".format(indeterminants[0], idx)
                    for idx in range(len(poly_like.indeterminants))
                ]
            remain |= indeterminants == poly_like.names

    if remain:
        return poly_like

    if indeterminants is None:
        return numpoly.polynomial(poly_like, dtype=dtype)
    return numpoly.polynomial(poly_like, indeterminants=indeterminants, dtype=dtype)



"""Monomial construction."""
from __future__ import division

import numpy
import numpoly


def monomial(indeterminants, start, stop=None, ordering="G", cross_truncation=1.):
    """
    Create an polynomial monomial expansion.

    Args:
        indeterminants (numpoly.ndpoly, str, Tuple[str, ...])
            The indeterminants used to create the monomials expansion.
        start (int, numpy.ndarray):
            The minimum polynomial to include. If int is provided, set as
            lowest total order. If array of int, set as lower order along each
            indeterminant.
        stop (int, numpy.ndarray):
            The maximum shape included. If omitted:
            ``stop <- start; start <- 0`` If int is provided, set as largest
            total order. If array of int, set as largest order along each
            indeterminant.
        ordering (str):
            The monomial ordering where the letters ``G``, ``I`` and ``R`` can
            be used to set grade, inverse and reverse to the ordering. For
            ``monomial(("x", "y"), start=0, stop=2, ordering=ordering)`` we
            get:

            ========  ==================
            ordering  output
            ========  ==================
            ""        [1 y y^2 x xy x^2]
            "G"       [1 y x y^2 xy x^2]
            "I"       [x^2 xy x y^2 y 1]
            "R"       [1 x x^2 y xy y^2]
            "GIR"     [y^2 xy x^2 y x 1]
            ========  ==================
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion.

    Returns:
        (numpoly.ndpoly):
            Monomial expansion.

    Examples:
        >>> print(numpoly.monomial("x", 4))
        [1 x x**2 x**3 x**4]
        >>> print(numpoly.monomial(("x", "y"), 4, 4, ordering="GR"))
        [x**4 x**3*y y**4 x**2*y**2 x*y**3]
        >>> print(numpoly.monomial(("x", "y"), [1, 1], [2, 2]))
        [x*y x*y**2 x**2*y x**2*y**2]

    """
    if stop is None:
        start, stop = numpy.array(0), start

    start = numpy.array(start, dtype=int)
    stop = numpy.array(stop, dtype=int)
    dimensions = max(start.size, stop.size, len(indeterminants))

    indices = bindex(
        start=numpy.min(start),
        stop=2*numpy.max(stop),
        dimensions=dimensions,
        ordering=ordering,
        cross_truncation=cross_truncation,
    )

    if start.size == 1:
        below = numpy.sum(indices, -1) >= start
    else:
        start = numpy.ones(dimensions, dtype=int)*start
        below = numpy.all(indices-start >= 0, -1)

    if stop.size == 1:
        above = numpy.sum(indices, -1) <= stop.item()
    else:
        stop = numpy.ones(dimensions, dtype=int)*stop
        above = numpy.all(stop-indices >= 0, -1)
    indices = indices[above*below]

    poly = numpoly.ndpoly(
        exponents=indices,
        shape=(len(indices),),
        indeterminants=indeterminants,
    )
    for coeff, key in zip(
            numpy.eye(len(indices), dtype=int), poly.keys):
        poly[key] = coeff
    return poly


def bindex(start, stop=None, dimensions=1, ordering="G", cross_truncation=1.):
    """
    Generate multi-indices for the monomial exponents.

    Args:
        start (Union[int, numpy.ndarray]):
            The lower order of the indices. If array of int, counts as lower
            bound for each axis.
        stop (Union[int, numpy.ndarray, None]):
            The maximum shape included. If omitted: stop <- start; start <- 0
            If int is provided, set as largest total order. If array of int,
            set as upper bound for each axis.
        dimensions (int):
            The number of dimensions in the expansion
        ordering (str):
            Criteria to sort the indices by.
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion. Ignored if ``stop`` is a array.

    Returns:
        list:
            Order list of indices.

    Examples:
        >>> print(bindex(2, 3, 2).tolist())
        [[0, 2], [1, 1], [2, 0], [0, 3], [1, 2], [2, 1], [3, 0]]
        >>> print(bindex(2, [1, 3], 2, cross_truncation=0).tolist())
        [[0, 2], [1, 1], [0, 3], [1, 2], [1, 3]]
        >>> print(bindex([1, 2], [2, 3], 2, cross_truncation=0).tolist())
        [[1, 2], [1, 3], [2, 2], [2, 3]]
        >>> print(bindex(1, 3, 2, cross_truncation=0).tolist())  # doctest: +NORMALIZE_WHITESPACE
        [[0, 1], [1, 0], [0, 2], [1, 1], [2, 0], [0, 3], [1, 2], [2, 1],
            [3, 0], [1, 3], [2, 2], [3, 1], [2, 3], [3, 2], [3, 3]]
        >>> print(bindex(1, 3, 2, cross_truncation=1).tolist())  # doctest: +NORMALIZE_WHITESPACE
        [[0, 1], [1, 0], [0, 2], [1, 1], [2, 0],
            [0, 3], [1, 2], [2, 1], [3, 0]]
        >>> print(bindex(1, 3, 2, cross_truncation=1.5).tolist())
        [[0, 1], [1, 0], [0, 2], [1, 1], [2, 0], [0, 3], [3, 0]]
        >>> print(bindex(1, 3, 2, cross_truncation=2).tolist())
        [[0, 1], [1, 0], [0, 2], [2, 0], [0, 3], [3, 0]]
        >>> print(bindex(0, 1, 3).tolist())
        [[0, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0]]
        >>> print(bindex(  # doctest: +NORMALIZE_WHITESPACE
        ...     [1, 1], 3, 2, cross_truncation=0).tolist())
        [[1, 1], [1, 2], [2, 1], [1, 3], [2, 2],
            [3, 1], [2, 3], [3, 2], [3, 3]]

    """
    if stop is None:
        start, stop = 0, start
    start = numpy.array(start, dtype=int).flatten()
    stop = numpy.array(stop, dtype=int).flatten()
    ordering = ordering.upper()
    start[start < 0] = 0

    indices = _bindex(start, stop, dimensions, cross_truncation)
    if "G" in ordering:
        indices = indices[numpy.argsort(numpy.sum(indices, -1))]

    if "I" in ordering:
        indices = indices[::-1]

    if "R" in ordering:
        indices = indices[:, ::-1]

    return indices


def _bindex(start, stop, dimensions=1, cross_truncation=1.):
    """Backend for the bindex function."""
    # At the beginning the current list of indices just ranges over the
    # last dimension.
    bound = stop.max()+1
    range_ = numpy.arange(bound, dtype=int)
    indices = range_[:, numpy.newaxis]

    for _ in range(dimensions-1):

        # Repeats the current set of indices.
        # e.g. [0,1,2] -> [0,1,2,0,1,2,...,0,1,2]
        indices = numpy.tile(indices, (bound, 1))

        # Stretches ranges over the new dimension.
        # e.g. [0,1,2] -> [0,0,...,0,1,1,...,1,2,2,...,2]
        front = range_.repeat(len(indices)//bound)[:, numpy.newaxis]

        # Puts them two together.
        indices = numpy.column_stack((front, indices))

        # Truncate at each iteration to ensure memory usage is low enough
        if stop.size == 1 and cross_truncation > 0:
            lhs = numpy.sum(indices**(1/cross_truncation), -1)
            rhs = numpy.max(stop, -1)**(1/cross_truncation)
            indices = indices[lhs <= rhs]
        else:
            indices = indices[numpy.all(indices <= stop, -1)]

    if start.size == 1:
        indices = indices[numpy.sum(indices, -1) >= start.item()]
    else:
        indices = indices[numpy.all(indices >= start, -1)]
    return indices.reshape(-1, dimensions)



"""Polynomial base class."""
from __future__ import division
from six import string_types

import numpy

from .exponent import exponents_to_keys, keys_to_exponents

from . import construct, array_function, poly_function


REDUCE_MAPPINGS = {
    numpy.add: numpy.sum,
    numpy.multiply: numpy.prod,
    numpy.logical_and: numpy.all,
    numpy.logical_or: numpy.any,
}
ACCUMULATE_MAPPINGS = {
    numpy.add: numpy.cumsum,
    numpy.multiply: numpy.cumprod,
}


class ndpoly(numpy.ndarray):  # pylint: disable=invalid-name
    """
    Polynomial as numpy array.

    An array object represents a multidimensional, homogeneous polynomial array
    of fixed-size items. An associated data-type object describes the format of
    each element in the array (its byte-order, how many bytes it occupies in
    memory, whether it is an integer, a floating point number, or something
    else, etc.)

    Arrays should be constructed using `polynomial`, `symbols` etc.

    Examples:
        >>> poly = ndpoly(
        ...     exponents=[(0, 1), (0, 0)], shape=(3,), indeterminants="x y")
        >>> poly["00"] = 1, 2, 3
        >>> poly["01"] = 4, 5, 6
        >>> print(numpy.array(list(poly.coefficients)))
        [[4 5 6]
         [1 2 3]]
        >>> print(poly)
        [4*y+1 5*y+2 6*y+3]
        >>> print(poly[0])
        4*y+1

    """

    # =================================================
    # Stuff to get subclassing of ndarray to run smooth
    # =================================================

    __array_priority__ = 16
    _dtype = None
    keys = None
    names = None

    def __new__(
            cls,
            exponents=((0,),),
            shape=(),
            indeterminants="z",
            dtype=None,
            **kwargs
    ):
        """
        Class constructor.

        Args:
            exponents (numpy.ndarray):
                The exponents in an integer array with shape ``(N, D)``, where
                ``N`` is the number of terms in the polynomial sum and ``D`` is
                the number of dimensions.
            shape (Tuple[int, ...]):
                Shape of created array.
            indeterminants (Union[str, Tuple[str], numpoly.ndpoly]):
                The name of the indeterminant variables in te polynomial. If
                polynomial, inherent from it. Else, pass argument to
                `numpoly.symbols` to create the indeterminants. If only one
                name is provided, but more than one is required, indeterminant
                will be extended with an integer index.
            dtype:
                Any object that can be interpreted as a numpy data type.
            kwargs:
                Extra arguments passed to `numpy.ndarray` constructor.

        """
        exponents = numpy.array(exponents, dtype=int)

        if isinstance(indeterminants, str):
            indeterminants = poly_function.symbols(indeterminants)
        if isinstance(indeterminants, ndpoly):
            indeterminants = indeterminants.names
        if len(indeterminants) == 1 and exponents.shape[1] > 1:
            indeterminants = tuple(
                "%s%d" % (str(indeterminants[0]), idx)
                for idx in range(exponents.shape[1])
            )

        keys = exponents_to_keys(exponents)

        dtype = int if dtype is None else dtype
        dtype_ = numpy.dtype([(key, dtype) for key in keys])

        obj = super(ndpoly, cls).__new__(
            cls, shape=shape, dtype=dtype_, **kwargs)
        obj._dtype = numpy.dtype(dtype)  # pylint: disable=protected-access
        obj.keys = keys  # pylint: disable=protected-access
        obj.names = tuple(indeterminants)  # pylint: disable=protected-access
        return obj

    def __array_finalize__(self, obj):
        """Finalize numpy constructor."""
        if obj is None:
            return
        self.keys = getattr(obj, "keys", None)
        self.names = getattr(obj, "names", None)
        self._dtype = getattr(obj, "_dtype", None)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Dispatch method for operators."""
        if method == "reduce":
            ufunc = REDUCE_MAPPINGS[ufunc]
        elif method == "accumulate":
            ufunc = ACCUMULATE_MAPPINGS[ufunc]
        else:
            assert method == "__call__", (
                "method %s not recognised" % method)
        assert ufunc in array_function.ARRAY_FUNCTIONS, (
            "function %s not supported by numpoly." % ufunc)
        return array_function.ARRAY_FUNCTIONS[ufunc](*inputs, **kwargs)

    def __array_function__(self, func, types, args, kwargs):
        """Dispatch method for functions."""
        assert func in array_function.ARRAY_FUNCTIONS, (
            "function %s not supported by numpoly." % func)
        return array_function.ARRAY_FUNCTIONS[func](*args, **kwargs)

    # ======================================
    # Properties specific for ndpoly objects
    # ======================================

    @property
    def coefficients(self):
        """
        Polynomial coefficients.

        Examples:
            >>> x, y = numpoly.symbols("x y")
            >>> poly = numpoly.polynomial([2*x**4, -3*y**2+14])
            >>> print(poly)
            [2*x**4 14-3*y**2]
            >>> print(numpy.array(poly.coefficients))
            [[ 0 14]
             [ 0 -3]
             [ 2  0]]

        """
        out = numpy.empty((len(self.keys),) + self.shape, dtype=self._dtype)
        for idx, key in enumerate(self.keys):
            out[idx] = numpy.ndarray.__getitem__(self, key)
        return list(out)

    @property
    def exponents(self):
        """
        Polynomial exponents.

        Examples:
            >>> x, y = numpoly.symbols("x y")
            >>> poly = numpoly.polynomial([2*x**4, -3*y**2+14])
            >>> print(poly)
            [2*x**4 14-3*y**2]
            >>> print(poly.exponents)
            [[0 0]
             [0 2]
             [4 0]]

        """
        return keys_to_exponents(self.keys)

    @staticmethod
    def from_attributes(
            exponents,
            coefficients,
            indeterminants="q",
            dtype=None,
            clean=True,
    ):
        """
        Construct polynomial from polynomial attributes.

        Args:
            exponents (numpy.ndarray):
                The exponents in an integer array with shape ``(N, D)``, where
                ``N`` is the number of terms in the polynomial sum and ``D`` is
                the number of dimensions.
            coefficients (Iterable[numpy.ndarray]):
                The polynomial coefficients. Must correspond to `exponents` by
                having the same length ``N``.
            indeterminants (Union[str, Tuple[str, ...], numpoly.ndpoly]):
                The indeterminants variables, either as string names or as
                simple polynomials. Must correspond to the exponents by having
                length ``D``.
            dtype (Optional[numpy.dtype]):
                The data type of the polynomial. If omitted, extract from
                `coefficients`.
            clean (bool):
                Clean up attributes, removing redundant indeterminants and
                exponents. Used to ensure alignment isn't broken.

        Examples:
            >>> numpoly.ndpoly.from_attributes(
            ...     exponents=[[0]], coefficients=[4])
            polynomial(4)
            >>> numpoly.ndpoly.from_attributes(
            ...     exponents=[[1]], coefficients=[[1, 2, 3]])
            polynomial([q, 2*q, 3*q])
            >>> numpoly.ndpoly.from_attributes(
            ...     exponents=[[0], [1]], coefficients=[[0, 1], [1, 1]])
            polynomial([q, 1+q])
            >>> numpoly.ndpoly.from_attributes(
            ...     exponents=[[0, 1], [1, 1]], coefficients=[[0, 1], [1, 1]])
            polynomial([q0*q1, q1+q0*q1])

        """
        return construct.polynomial_from_attributes(
            exponents=exponents,
            coefficients=coefficients,
            indeterminants=indeterminants,
            dtype=dtype,
            clean=clean,
        )

    @property
    def indeterminants(self):
        """
        Polynomial indeterminants.

        Examples:
            >>> x, y = numpoly.symbols("x y")
            >>> poly = numpoly.polynomial([2*x**4, -3*y**2+14])
            >>> print(poly)
            [2*x**4 14-3*y**2]
            >>> print(poly.indeterminants)
            [x y]

        """
        return construct.polynomial_from_attributes(
            exponents=numpy.eye(len(self.names), dtype=int),
            coefficients=numpy.eye(len(self.names), dtype=int),
            indeterminants=self.names,
        )

    def isconstant(self):
        """Check if a polynomial is constant or not."""
        return poly_function.isconstant(self)

    def todict(self):
        """
        Cast to dict where keys are exponents and values are coefficients.

        Examples:
            >>> x, y = numpoly.symbols("x y")
            >>> poly = 2*x**4-3*y**2+14
            >>> print(poly)
            14-3*y**2+2*x**4
            >>> print(poly.todict())
            {(0, 0): 14, (0, 2): -3, (4, 0): 2}

        """
        return {tuple(exponent): coefficient
                for exponent, coefficient in zip(
                    self.exponents, self.coefficients)}

    def toarray(self):
        """Cast polynomial to numpy.ndarray, if possible."""
        return poly_function.toarray(self)

    # =============================================
    # Override numpy properties to work with ndpoly
    # =============================================

    @property
    def dtype(self):
        """Show coefficient dtype instead of the structured array."""
        return self._dtype

    def astype(self, dtype, **kwargs):
        """Wrap ndarray.astype."""
        coefficients = [coefficient.astype(dtype, **kwargs)
                        for coefficient in self.coefficients]
        return construct.polynomial_from_attributes(
            self.exponents, coefficients, self.names)

    def round(self, decimals=0, out=None):
        """Wrap ndarray.round."""
        # Not sure why it is required. Likely a numpy bug.
        return array_function.around(self, decimals=decimals, out=out)

    # ============================================================
    # Override dunder methods that isn't dealt with by dispatching
    # ============================================================

    def __call__(self, *args, **kwargs):
        """Evaluate polynomial."""
        return poly_function.call(self, *args, **kwargs)

    def __eq__(self, other):
        """Left equality."""
        return array_function.equal(self, other)

    def __getitem__(self, index):
        """
        Get array item or slice.

        Args:
            index (Union[int, str, Tuple[int, ...], numpy.ndarray]):
                The index to extract. If string type, extract using th
                underlying structured array data type. Else, extract as
                numpy.ndarray.

        Examples:
            >>> x, y = numpoly.symbols("x y")
            >>> poly = numpoly.polynomial([[1-4*x, x**2], [y-3, x*y*y]])
            >>> print(poly)
            [[1-4*x x**2]
             [-3+y x*y**2]]
            >>> print(poly[0])
            [1-4*x x**2]
            >>> print(poly[:, 1])
            [x**2 x*y**2]
            >>> print(poly["10"])
            [[-4  0]
             [ 0  0]]

        """
        if isinstance(index, string_types):
            return numpy.asarray(super(ndpoly, self).__getitem__(index))
        return construct.polynomial_from_attributes(
            exponents=self.exponents,
            coefficients=[coeff[index] for coeff in self.coefficients],
            indeterminants=self.names,
        )

    def __iter__(self):
        """Iterate polynomial array."""
        coefficients = numpy.array(list(self.coefficients))
        return iter(construct.polynomial_from_attributes(
            exponents=self.exponents,
            coefficients=coefficients[:, idx],
            indeterminants=self.names,
        ) for idx in range(len(self)))

    def __ne__(self, other):
        """Not equal."""
        return array_function.not_equal(self, other)

    def __repr__(self):
        """Canonical string representation."""
        return array_function.array_repr(self)

    def __str__(self):
        """Pretty string representation."""
        return array_function.array_str(self)



"""View inputs as arrays with at least two dimensions."""
import numpy
import numpoly

from .common import implements


@implements(numpy.atleast_2d)
def atleast_2d(*arys):
    """
    View inputs as arrays with at least two dimensions.

    Args:
        arys (numpoly.ndpoly):
            One or more array-like sequences. Non-array inputs are converted
            to arrays. Arrays that already have two or more dimensions are
            preserved.

    Returns:
        (numpoly.ndpoly):
            An array, or list of arrays, each with ``a.ndim >= 2``. Copies are
            avoided where possible, and views with two or more dimensions are
            returned.

    Examples:
        >>> numpoly.atleast_2d(numpoly.symbols("x"))
        polynomial([[x]])
        >>> numpoly.atleast_2d(1, [2, 3], [[4]])
        [polynomial([[1]]), polynomial([[2, 3]]), polynomial([[4]])]

    """
    arys = [numpoly.aspolynomial(ary) for ary in arys]
    arys = [ary if ary.ndim else ary.reshape(1, 1) for ary in arys]
    arys = [ary[numpy.newaxis] if ary.ndim == 1 else ary for ary in arys]
    if len(arys) == 1:
        return arys[0]
    return arys



"""Calculate the absolute value element-wise."""
import numpy
import numpoly

from .common import implements, simple_dispatch


@implements(numpy.abs, numpy.absolute)
def absolute(x, out=None, where=True, **kwargs):
    r"""
    Calculate the absolute value element-wise.

    Args:
        x (numpoly.ndpoly):
            Input array.
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Union[bool, numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        absolute (numpoly.ndpoly):
            An ndarray containing the absolute value of. each element in `x`.
            For complex input, ``a + ib``, the absolute value is
            :math:`\sqrt{a^2+b^2}`. This is a scalar if `x` is a scalar.

    Examples:
        >>> x = numpoly.symbols("x")
        >>> poly = numpoly.polynomial([-1.2, 1.2, -2.3*x, 2.3*x])
        >>> poly
        polynomial([-1.2, 1.2, -2.3*x, 2.3*x])
        >>> numpoly.absolute(poly)
        polynomial([1.2, 1.2, 2.3*x, 2.3*x])
        >>> poly = numpoly.polynomial([x, 1j*x, (3+4j)*x])
        >>> poly
        polynomial([x, 1j*x, (3+4j)*x])
        >>> numpoly.absolute(poly)
        polynomial([x, x, 5.0*x])

    """
    return simple_dispatch(
        numpy_func=numpy.absolute,
        inputs=(x,),
        out=out,
        where=where,
        **kwargs
    )



"""Join a sequence of arrays along an existing axis."""
import numpy
import numpoly

from .common import implements


@implements(numpy.concatenate)
def concatenate(arrays, axis=0, out=None):
    """
    Join a sequence of arrays along an existing axis.

    Args:
        arrays (Iterable[numpoly.ndpoly]):
            The arrays must have the same shape, except in the dimension
            corresponding to `axis` (the first, by default).
        axis (Optional[int]):
            The axis along which the arrays will be joined.  If axis is None,
            arrays are flattened before use.  Default is 0.
        out (Optional[numpy.ndarray]):
            If provided, the destination to place the result. The shape must be
            correct, matching that of what concatenate would have returned if
            no out argument were specified.

    Returns:
        (numpoly.ndpoly):
            The concatenated array.

    Examples:
        >>> a = numpy.array([[1, 2], [3, 4]])
        >>> b = numpoly.symbols("x y").reshape(1, 2)
        >>> numpoly.concatenate((a, b), axis=0)
        polynomial([[1, 2],
                    [3, 4],
                    [x, y]])
        >>> numpoly.concatenate((a, b.T), axis=1)
        polynomial([[1, 2, x],
                    [3, 4, y]])
        >>> numpoly.concatenate((a, b), axis=None)
        polynomial([1, 2, 3, 4, x, y])

    """
    arrays = numpoly.align_indeterminants(*arrays)
    collections = [arg.todict() for arg in arrays]

    output = {}
    keys = {arg for collection in collections for arg in collection}
    for key in keys:
        values = [(collection[key] if key in collection
                   else numpy.zeros(array.shape, dtype=bool))
                  for collection, array in zip(collections, arrays)]
        output[key] = numpy.concatenate(values, axis=axis, out=out)

    exponents = sorted(output)
    coefficients = [output[exponent] for exponent in exponents]
    return numpoly.ndpoly.from_attributes(
        exponents=exponents,
        coefficients=coefficients,
        indeterminants=arrays[0].indeterminants,
    )



"""Cast polynomial to numpy.ndarray, if possible."""
import numpy
import numpoly


def toarray(poly):
    """
    Cast polynomial to numpy.ndarray, if possible.

    Args:
        poly (numpoly.ndpoly):
            polynomial to cast.

    Returns:
        (numpy.ndarray):
            Numpy array.

    Raises:
        ValueError:
            Only constant polynomials can be cast to numpy.ndarray.

    Examples:
        >>> numpoly.polynomial([1, 2]).toarray()
        array([1, 2])
        >>> numpoly.toarray(numpoly.symbols("x"))
        Traceback (most recent call last):
            ...
        ValueError: only constant polynomials can be converted to array.

    """
    poly = numpoly.aspolynomial(poly)
    if not poly.isconstant():
        raise ValueError(
            "only constant polynomials can be converted to array.")
    return numpy.array(poly.coefficients[0])



"""Polynomials differentiation functions."""
import numpy
import numpoly


def diff(poly, *diffvars):
    """
    Polynomial differential operator.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to differentiate.
        diffvars (numpoly.ndpoly, str):
            Singleton variables to take derivative off.

    Returns:
        Same as ``poly`` but differentiated with respect to ``diffvars``.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> poly = numpoly.polynomial([1, x, x*y**2+1])
        >>> print(poly)
        [1 x 1+x*y**2]
        >>> print(numpoly.diff(poly, "x"))
        [0 1 y**2]
        >>> print(numpoly.diff(poly, x, y))
        [0 0 2*y]

    """
    poly = numpoly.polynomial(poly)

    for diffvar in diffvars:
        if isinstance(diffvar, str):
            idx = poly.names.index(diffvar)
        elif isinstance(diffvar, int):
            idx = diffvar
        else:
            diffvar = numpoly.polynomial(diffvar)
            assert len(diffvar.names) == 1, "only one at the time"
            assert numpy.all(diffvar.exponents == 1), (
                "derivative variable assumes singletons")
            idx = poly.names.index(diffvar.names[0])

        exponents = poly.exponents
        coefficients = [
            (exponent[idx]*coefficient.T).T
            for exponent, coefficient in zip(exponents, poly.coefficients)
        ]
        exponents[:, idx] -= 1
        if numpy.any(exponents < 0):
            indices = exponents[:, idx] == -1
            coefficients = [coefficient for coefficient, idx in zip(
                coefficients, indices) if not idx]
            exponents = numpy.delete(
                exponents, numpy.argwhere(indices), axis=0)

        poly = numpoly.ndpoly.from_attributes(
            exponents=exponents,
            coefficients=coefficients,
            indeterminants=poly.names,
        )
    return poly


def gradient(poly):
    """
    Polynomial gradient operator.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to differentiate. If polynomial vector is provided,
            the Jacobi-matrix is returned instead.

    Returns:
        Same as ``poly`` but with an extra first dimensions, one for each
        variable in ``poly.indeterminants``, filled with gradient values.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> poly = 5*x**5+4
        >>> print(numpoly.gradient(poly))
        [25*x**4]
        >>> poly = 4*x**3+2*y**2+3
        >>> print(numpoly.gradient(poly))
        [12*x**2 4*y]
        >>> poly = numpoly.polynomial([1, x**3, x*y**2+1])
        >>> print(numpoly.gradient(poly))
        [[0 3*x**2 y**2]
         [0 0 2*x*y]]

    """
    polys = [diff(poly, diffvar)[numpy.newaxis]
             for diffvar in poly.indeterminants]
    return numpoly.concatenate(polys, axis=0)


def hessian(poly):
    """
    Construct Hessian matrix of polynomials.

    Make Hessian matrix out of a polynomial. Tensor is returned if polynomial
    is a vector.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to differentiate.

    Returns:
        Same as ``poly`` but with two extra dimensions, one for each
        variable in ``poly.indeterminants``, filled with Hessian values.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> poly = 5*x**5+4
        >>> print(numpoly.hessian(poly))
        [[100*x**3]]
        >>> poly = 4*x**3+2*y**2+3
        >>> print(numpoly.hessian(poly))
        [[24*x 0]
         [0 4]]
        >>> poly = numpoly.polynomial([1, x, x*y**2+1])
        >>> print(numpoly.hessian(poly))
        [[[0 0 0]
          [0 0 2*y]]
        <BLANKLINE>
         [[0 0 2*y]
          [0 0 2*x]]]

    """
    return gradient(gradient(poly))



"""Clean up polynomial attributes."""
import numpy
import numpoly


def clean_attributes(poly):
    """Clean up polynomial attributes."""
    return numpoly.ndpoly.from_attributes(
        exponents=poly.exponents,
        coefficients=poly.coefficients,
        indeterminants=poly.names,
        dtype=poly.dtype,
        clean=True,
    )


def postprocess_attributes(exponents, coefficients, indeterminants):
    """Clean up polynomial attributes."""
    coefficients = [numpy.asarray(coefficient) for coefficient in coefficients]
    pairs = [
        (exponent, coefficient)
        for exponent, coefficient in zip(exponents, coefficients)
        if numpy.any(coefficient) or not any(exponent)
    ]
    if pairs:
        exponents, coefficients = zip(*pairs)
    else:
        exponents = [(0,)*len(indeterminants)]
        coefficients = numpy.zeros(
            (1,)+coefficients[0].shape, dtype=coefficients[0].dtype)

    exponents = numpy.asarray(exponents, dtype=int)
    if isinstance(indeterminants, numpoly.ndpoly):
        indeterminants = indeterminants.names

    if isinstance(indeterminants, str):
        if exponents.shape[1] > 1:
            indeterminants = ["%s%d" % (indeterminants, idx)
                              for idx in range(exponents.shape[1])]
        else:
            indeterminants = [indeterminants]
    assert len(indeterminants) == exponents.shape[1], (indeterminants, exponents)

    indices = numpy.any(exponents != 0, 0)
    assert exponents.shape[1] == len(indices), (exponents, indices)
    if not numpy.any(indices):
        indices[0] = True
    exponents = exponents[:, indices]
    assert exponents.size, (exponents, indices)
    indeterminants = numpy.array(indeterminants)[indices].tolist()

    assert len(exponents.shape) == 2, exponents
    assert len(exponents) == len(coefficients)
    assert len(numpy.unique(exponents, axis=0)) == exponents.shape[0], exponents
    assert sorted(set(indeterminants)) == sorted(indeterminants)

    return exponents, coefficients, indeterminants



"""Return (x1 == x2) element-wise."""
import numpy
import numpoly

from .common import implements, simple_dispatch


@implements(numpy.equal)
def equal(x1, x2, out=None, where=True, **kwargs):
    """
    Return (x1 == x2) element-wise.

    Args:
        x1, x2 (numpoly.ndpoly):
            Input arrays. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value.
            Note that if an uninitialized `out` array is created via the
            default ``out=None``, locations within it where the condition is
            False will remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (Union[numpy.ndarray, numpy.generic]):
            Output array, element-wise comparison of `x1` and `x2`. Typically
            of type bool, unless ``dtype=object`` is passed. This is a scalar
            if both `x1` and `x2` are scalars.

    Examples:
        >>> x, y, z = xyz = numpoly.symbols("x y z")
        >>> numpoly.equal(xyz, x)
        array([ True, False, False])
        >>> numpoly.equal(xyz, [y, y, z])
        array([False,  True,  True])

    """
    x1, x2 = numpoly.align_polynomials(x1, x2)
    if out is None:
        out = numpy.ones(x1.shape, dtype=bool)
    collection = x1.todict()
    for exponent, coefficient in x2.todict().items():
        if exponent in collection:
            out &= numpy.equal(
                collection.pop(exponent), coefficient, where=where, **kwargs)
        else:
            out &= coefficient == 0
    for _, coefficient in collection.items():
        out &= coefficient == 0
    return out



"""Polynomial functionality."""
from .call import call
from .derivative import diff, gradient, hessian
from .isconstant import isconstant
from .monomial import monomial
from .symbols import symbols
from .toarray import toarray



"""Construct polynomial from polynomial attributes."""
import numpoly

from . import clean as clean_


def polynomial_from_attributes(
        exponents,
        coefficients,
        indeterminants,
        dtype=None,
        clean=True,
):
    """
    Construct polynomial from polynomial attributes.

    Args:
        exponents (numpy.ndarray):
            The exponents in an integer array with shape ``(N, D)``, where
            ``N`` is the number of terms in the polynomial sum and ``D`` is
            the number of dimensions.
        coefficients (Iterable[numpy.ndarray]):
            The polynomial coefficients. Must correspond to `exponents` by
            having the same length ``N``.
        indeterminants (Union[str, Tuple[str, ...], numpoly.ndpoly]):
            The indeterminants variables, either as string names or as
            simple polynomials. Must correspond to the exponents by having
            length ``D``.
        dtype (Optional[numpy.dtype]):
            The data type of the polynomial. If omitted, extract from
            `coefficients`.
        clean (bool):
            Clean up attributes, removing redundant indeterminants and
            exponents.

    Examples:
        (numpoly.ndpoly):
            Polynomial array with attributes determined by the input.

    """
    if clean:
        exponents, coefficients, indeterminants = clean_.postprocess_attributes(
            exponents, coefficients, indeterminants)
    dtype = coefficients[0].dtype if dtype is None else dtype
    poly = numpoly.ndpoly(
        exponents=exponents,
        shape=coefficients[0].shape,
        indeterminants=indeterminants,
        dtype=dtype,
    )
    for exponent, values in zip(poly.keys, coefficients):
        poly[exponent] = values
    return poly



"""Construct symbol variables."""
import re
import string

import numpy
import numpoly


def symbols(names="q", asarray=False, dtype="i8"):
    """
    Construct symbol variables.

    Most directly be providing a list of string names. But a set of shorthands
    also exists:

    * ``,`` and `` `` (space) can be used as a variable delimiter.
    * ``{number}:{number}`` can be used to define a numerical range.
    * ``{letter}:{letter}`` can be used to define a alphabet range.

    Args:
        names (str, Tuple[str, ...]):
            Indeterminants are determined by splitting the string on space. If
            iterable of strings, indeterminants defined directly.
        asarray (bool):
            Enforce output as array even in the case where there is only one
            variable.
        dtype (numpy.dtype):
            The data type of the polynomial coefficients.

    Returns:
        (numpoly.ndpoly):
            Polynomial array with unit components in each dimension.

    Examples:
        >>> print(numpoly.symbols("q"))
        q
        >>> print(numpoly.symbols("q,"))
        [q]
        >>> print(numpoly.symbols("q", asarray=True))
        [q]
        >>> print(numpoly.symbols(["alpha", "beta"]))
        [alpha beta]
        >>> print(numpoly.symbols("x y z"))
        [x y z]
        >>> print(numpoly.symbols("a,b,c"))
        [a b c]
        >>> print(numpoly.symbols("q:7"))
        [q0 q1 q2 q3 q4 q5 q6]
        >>> print(numpoly.symbols("q3:6"))
        [q3 q4 q5]
        >>> print(numpoly.symbols("za:f"))
        [za zb zc zd ze zf]

    """
    if not isinstance(names, str):
        names = list(names)

    else:
        names = re.sub(" ", ",", names)
        if "," in names:
            asarray = True
            names = [name for name in names.split(",") if name]

        elif re.search(r"(\d*):(\d+)", names):

            match = re.search(r"(\d*):(\d+)", names)
            start = int(match.group(1) or 0)
            end = int(match.group(2))
            names = [
                names.replace(match.group(0), str(idx))
                for idx in range(start, end)
            ]

        elif re.search(r"([a-zA-Z]+):([a-zA-Z]+)", names):

            match = re.search(r"([a-zA-Z]):([a-zA-Z])", names)
            start = string.ascii_letters.index(match.group(1))
            end = string.ascii_letters.index(match.group(2))
            names = [
                names.replace(
                    match.group(0), string.ascii_letters[idy])
                for idy in range(start, end+1)
            ]

    exponents = numpy.eye(len(names), dtype=int)
    coefficients = numpy.eye(len(names), dtype=dtype)
    if len(names) == 1 and not asarray:
        coefficients = coefficients[0]
    return numpoly.ndpoly.from_attributes(
        exponents=exponents,
        coefficients=coefficients,
        indeterminants=names,
    )



"""Attempt to cast an object into a polynomial array."""
import numpy
import numpoly

from .compose import compose_polynomial_array


def polynomial(
        poly_like=None,
        indeterminants="q",
        dtype=None,
):
    """
    Attempt to cast an object into a polynomial array.

    Supports various casting options:

    ``dict``
        Keys are tuples that represent polynomial exponents, and values are
        numpy arrays that represents polynomial coefficients.
    ``numpoly.ndpoly``
        Copy of the polynomial.
    ``int``, ``float``, ``numpy.ndarray``
        Constant term polynomial.
    ``sympy.Poly``
        Convert polynomial from ``sympy`` to ``numpoly``, if possible.
    ``Iterable``
        Multivariate array construction.
    ``numpy structured array``
        Assumes that the raw polynomial core. Used for developer
        convenience.

    Args:
        poly_like (Any):
            Input to be converted to a `numpoly.ndpoly` polynomial type.
        indeterminants (str, Tuple[str, ...]):
            Name of the indeterminant variables. If possible to infer from
            ``poly_like``, this argument will be ignored.
        dtype (type, numpy.dtype):
            Data type used for the polynomial coefficients.

    Returns:
        (numpoly.ndpoly):
            Polynomial based on input ``poly_like``.

    Examples:
        >>> print(numpoly.polynomial({(1,): 1}))
        q
        >>> x, y = numpoly.symbols("x y")
        >>> print(x**2 + x*y + 2)
        2+x*y+x**2
        >>> poly = -3*x + x**2 + y
        >>> print(numpoly.polynomial([x*y, x, y]))
        [x*y x y]
        >>> print(numpoly.polynomial([1, 2, 3]))
        [1 2 3]
        >>> import sympy
        >>> x_, y_ = sympy.symbols("x, y")
        >>> print(numpoly.polynomial(3*x_*y_ - 4 + x_**5))
        x**5+3*x*y-4

    """
    if poly_like is None:
        poly = numpoly.ndpoly(
            exponents=[(0,)],
            shape=(),
            indeterminants=indeterminants,
            dtype=dtype,
        )
        poly["0"] = 0

    elif isinstance(poly_like, dict):
        poly = numpoly.ndpoly(exponents=[(0,)], shape=())
        exponents, coefficients = zip(*list(poly_like.items()))
        poly = numpoly.ndpoly.from_attributes(
            exponents=exponents,
            coefficients=coefficients,
            indeterminants=indeterminants,
            dtype=dtype,
        )

    elif isinstance(poly_like, numpoly.ndpoly):
        if dtype and poly_like.dtype != dtype:
            poly = poly_like.astype(dtype)
        else:
            poly = poly_like.copy()

    # assume polynomial converted to structured array
    elif isinstance(poly_like, numpy.ndarray) and poly_like.dtype.names:
        exponents = numpoly.keys_to_exponents(poly_like.dtype.names)
        coefficients = [poly_like[key] for key in poly_like.dtype.names]
        poly = numpoly.ndpoly.from_attributes(
            exponents=exponents,
            coefficients=coefficients,
            indeterminants=indeterminants,
        )

    elif isinstance(poly_like, (int, float, numpy.ndarray, numpy.generic)):
        poly = numpoly.ndpoly.from_attributes(
            exponents=[(0,)],
            coefficients=numpy.array([poly_like]),
            indeterminants=indeterminants,
            dtype=dtype,
        )

    # handler for sympy objects
    elif hasattr(poly_like, "as_poly"):
        poly_like = poly_like.as_poly()
        exponents = poly_like.monoms()
        coefficients = [int(coeff) if coeff.is_integer else float(coeff)
                        for coeff in poly_like.coeffs()]
        indeterminants = [str(elem) for elem in poly_like.gens]
        poly = numpoly.ndpoly.from_attributes(
            exponents=exponents,
            coefficients=coefficients,
            indeterminants=indeterminants,
        )

    else:
        poly = compose_polynomial_array(
            arrays=poly_like,
            dtype=dtype,
        )

    return poly



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['numpoly',
 'numpoly.array_function',
 'numpoly.construct',
 'numpoly.poly_function']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.16,<2.0']

setup_kwargs = {
    'name': 'numpoly',
    'version': '0.0.13',
    'description': 'Polynomials as a numpy datatype',
    'long_description': 'Numpoly is a generic library for creating, manipulating polynomial arrays.\n\nMany numerical analysis, prominent in for example uncertainty quantification,\nuses polynomial approximations as proxy for real models to do analysis on.\nThese models are often solutions to non-linear problems discretized with high\nmesh. As such, the corresponding polynomial approximation consist of high\nnumber of dimensions and large multi-dimensional polynomial coefficients.\n\n``numpoly`` is a subclass of ``numpy.ndarray`` implemented to represent\npolynomials as array element. As such is fast and scales very well with the\nsize of the coefficients. It is also compatible with most ``numpy`` functions,\nwhere that makes sense, making the interface fairly intuitive. Some of the\ninterface is also inspired by the ``sympy`` interface.\n\n|circleci| |codecov| |pypi|\n\n.. |circleci| image:: https://circleci.com/gh/jonathf/numpoly/tree/master.svg?style=shield\n    :target: https://circleci.com/gh/jonathf/numpoly/tree/master\n.. |codecov| image:: https://codecov.io/gh/jonathf/numpoly/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/jonathf/numpoly\n.. |pypi| image:: https://badge.fury.io/py/numpoly.svg\n    :target: https://badge.fury.io/py/numpoly\n\n.. contents:: Table of Contents:\n\nInstallation\n------------\n\nInstallation should be straight forward:\n\n.. code-block:: bash\n\n    pip install numpoly\n\nAnd you should be ready to go.\n\nExample usage\n-------------\n\nConstructing polynomial is typically done using one of the available\nconstructors:\n\n.. code-block:: python\n\n   >>> poly1 = numpoly.monomial(("x", "y"), start=0, stop=3)\n   >>> print(poly1)\n   [1 y x x*y x**2 y**2 y**3 x*y**2 x**2*y x**3]\n\nIt is also possible to construct your own from symbols:\n\n.. code-block:: python\n\n   >>> x, y = numpoly.symbols("x y")\n   >>> poly2 = numpoly.polynomial([1, x**2-1, x*y, y**2-1])\n   >>> print(poly2)\n   [1 -1+x**2 x*y -1+y**2]\n\nOr in combination with other numpy objects:\n\n.. code-block:: python\n\n   >>> poly3 = x**numpy.arange(4)-y**numpy.arange(3, -1, -1)\n   >>> print(poly3)\n   [1-y**3 -y**2+x -y+x**2 -1+x**3]\n\nThe polynomials can be evaluated as needed:\n\n.. code-block:: python\n\n   >>> print(poly1(1, 2))\n   [1 2 1 2 1 4 8 4 2 1]\n   >>> print(poly2(x=[1, 2]))\n   [[1 1]\n    [0 3]\n    [y 2*y]\n    [-1+y**2 -1+y**2]]\n   >>> print(poly1(x=y, y=2*x))\n   [1 2*x y 2*x*y y**2 4*x**2 8*x**3 4*x**2*y 2*x*y**2 y**3]\n\nThe polynomials also support many numpy operations:\n\n.. code-block:: python\n\n   >>> print(numpy.reshape(poly2, (2, 2)))\n   [[1 -1+x**2]\n    [x*y -1+y**2]]\n   >>> print(poly1[::3].astype(float))\n   [1.0 x*y y**3 x**3]\n   >>> print(numpy.sum(poly1.reshape(2, 5), 0))\n   [1+y**2 y+y**3 x+x*y**2 x*y+x**2*y x**2+x**3]\n\nThere are also several polynomial specific operators:\n\n.. code-block:: python\n\n   >>> print(numpoly.diff(poly3, y))\n   [-3*y**2 -2*y -1 0]\n   >>> print(numpoly.gradient(poly3))\n   [[0 1 2*x 3*x**2]\n    [-3*y**2 -2*y -1 0]]\n\n\nDevelopment\n-----------\n\nDevelopment is done using `Poetry <https://poetry.eustace.io/>`_ manager.\nInside the repository directory, install and create a virtual enviroment with:\n\n.. code-block:: bash\n\n   poetry install\n\nTo run tests, run:\n\n.. code-block:: bash\n\n   poentry run pytest numpoly test --doctest-modules\n\nQuestions & Troubleshooting\n---------------------------\n\nFor any problems and questions you might have related to ``numpoly``, please\nfeel free to file an `issue <https://github.com/jonathf/numpoly/issues>`_.\n',
    'author': 'Jonathan Feinberg',
    'author_email': 'jonathf@gmail.com',
    'url': 'https://github.com/jonathf/numpoly',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
}


setup(**setup_kwargs)



"""Return the floor of the input, element-wise."""
import numpy
import numpoly

from .common import implements, simple_dispatch


@implements(numpy.floor)
def floor(x, out=None, where=True, **kwargs):
    r"""
    Return the floor of the input, element-wise.

    The floor of the scalar `x` is the largest integer `i`, such that
    `i <= x`.  It is often denoted as :math:`\lfloor x \rfloor`.

    Args:
        x (numpoly.ndpoly):
            Input data.
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Union[bool, numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        y (numpoly.ndpoly):
            The floor of each element in `x`. This is a scalar if `x` is
            a scalar.

    Examples:
        >>> x = numpoly.symbols("x")
        >>> numpoly.floor([-1.7*x, x-1.5, -0.2, 3.2+1.5*x, 1.7, 2.0])
        polynomial([-2.0*x, -2.0+x, -1.0, 3.0+x, 1.0, 2.0])

    """
    return simple_dispatch(
        numpy_func=numpy.floor,
        inputs=(x,),
        out=out,
        where=where,
        **kwargs
    )



"""Return a scalar type which is common to the input arrays."""
import numpy
import numpoly

from .common import implements


@implements(numpy.common_type)
def common_type(*arrays):
    """
    Return a scalar type which is common to the input arrays.

    The return type will always be an inexact (i.e. floating point) scalar
    type, even if all the arrays are integer arrays. If one of the inputs is an
    integer array, the minimum precision type that is returned is a 64-bit
    floating point dtype.

    All input arrays except int64 and uint64 can be safely cast to the
    returned dtype without loss of information.

    Args:
        arrays (numpoly.ndpoly):
            Input arrays.

    Return:
        out (numpy.generic):
            Data type code.

    Examples:
        >>> numpoly.common_type(numpy.array(2, dtype=numpy.float32)).__name__
        'float32'
        >>> numpoly.common_type(numpoly.symbols("x")).__name__
        'float64'
        >>> numpoly.common_type(
        ...     numpy.arange(3), 1j*numpoly.symbols("x"), 45).__name__
        'complex128'

    """
    arrays = [numpoly.aspolynomial(array) for array in arrays]
    arrays = [array[array.keys[0]] for array in arrays]
    return numpy.common_type(*arrays)



"""Mapping between polynomial exponents and their string counterpart."""
from __future__ import unicode_literals
from string import printable  # pylint: disable=no-name-in-module

import numpy

FORWARD_DICT = dict(enumerate(numpy.array(list(printable), dtype="U1")))
FORWARD_MAP = numpy.vectorize(FORWARD_DICT.get)
INVERSE_DICT = {value: key for key, value in FORWARD_DICT.items()}
INVERSE_MAP = numpy.vectorize(INVERSE_DICT.get)


def keys_to_exponents(keys):
    """
    Convert string keys into numerical exponents.

    Args:
        keys (Union[str, Iterable[str]]):
            String values to convert to exponents.

    Returns:
        (numpy.ndarray):
            Exponent values.

    Examples:
        >>> keys_to_exponents("0P")
        array([ 0, 51])
        >>> keys_to_exponents(["12", "21", "0P"])
        array([[ 1,  2],
               [ 2,  1],
               [ 0, 51]])

    """
    keys = numpy.asarray(keys, dtype="U")
    if not keys.shape:
        return numpy.array([INVERSE_DICT[key] for key in keys.item()])
    keys = keys.view("U1").reshape(len(keys), -1)
    assert "" not in keys, "all keys have to be the same length"
    exponents = INVERSE_MAP(keys)
    return exponents


def exponents_to_keys(exponents):
    """
    Convert exponent values into string equivalent.

    Args:
        exponents (numpy.numpy):
            Exponent values.
            String values to convert to exponents.

    Returns:
        (numpy.ndarray):
            String representation of exponent values.

    Examples:
        >>> exponents_to_keys([0, 1, 2]) == u"012"
        True
        >>> exponents_to_keys([[0, 0], [0, 1], [9, 44]]) == [u"00", u"01", u"9I"]
        array([ True,  True,  True])

    """
    exponents = numpy.asarray(exponents, dtype=int)
    if len(exponents.shape) == 1:
        return exponents_to_keys(exponents.reshape(1, -1))[0]
    assert len(exponents.shape) == 2, "invalid exponent input"
    keys = FORWARD_MAP(exponents).flatten()
    keys = numpy.array(keys.view("U%d" % exponents.shape[-1]))
    return keys



"""Align polynomials."""
import numpy
import numpoly


def align_polynomials(*polys):
    """
    Align polynomial such that dimensionality, shape, etc. are compatible.

    Args:
        polys (numpoly.ndpoly):
            Polynomial to make adjustment to.

    Returns:
        (Tuple[numpoly.ndpoly, ...]):
            Same as ``polys``, but internal adjustments made to make them
            compatible for further operations.

    """
    polys = align_shape(*polys)
    polys = align_indeterminants(*polys)
    polys = align_exponents(*polys)
    polys = align_dtype(*polys)
    return polys


def align_shape(*polys):
    """
    Align polynomial by shape.

    Args:
        polys (numpoly.ndpoly):
            Polynomial to make adjustment to.

    Returns:
        (Tuple[numpoly.ndpoly, ...]):
            Same as ``polys``, but internal adjustments made to make them
            compatible for further operations.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> poly1 = 4*x
        >>> poly2 = numpoly.polynomial([[2*x+1, 3*x-y]])
        >>> print(poly1.shape)
        ()
        >>> print(poly2.shape)
        (1, 2)
        >>> poly1, poly2 = numpoly.align_shape(poly1, poly2)
        >>> print(poly1)
        [[4*x 4*x]]
        >>> print(poly2)
        [[1+2*x -y+3*x]]
        >>> print(poly1.shape)
        (1, 2)
        >>> print(poly2.shape)
        (1, 2)

    """
    polys = [numpoly.aspolynomial(poly) for poly in polys]
    common = 1
    for poly in polys:
        common = numpy.ones(poly.coefficients[0].shape, dtype=int)*common

    polys = [poly.from_attributes(
        exponents=poly.exponents,
        coefficients=[coeff*common for coeff in poly.coefficients],
        indeterminants=poly.indeterminants,
    ) for poly in polys]
    assert numpy.all(common.shape == poly.shape for poly in polys)
    return tuple(polys)


def align_indeterminants(*polys):
    """
    Align polynomial by indeterminants.

    Args:
        polys (numpoly.ndpoly):
            Polynomial to make adjustment to.

    Returns:
        (Tuple[numpoly.ndpoly, ...]):
            Same as ``polys``, but internal adjustments made to make them
            compatible for further operations.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> poly1, poly2 = numpoly.polynomial([2*x+1, 3*x-y])
        >>> print(poly1.indeterminants)
        [x]
        >>> print(poly2.indeterminants)
        [x y]
        >>> poly1, poly2 = numpoly.align_indeterminants(poly1, poly2)
        >>> print(poly1)
        1+2*x
        >>> print(poly2)
        -y+3*x
        >>> print(poly1.indeterminants)
        [x y]
        >>> print(poly2.indeterminants)
        [x y]

    """
    polys = [numpoly.aspolynomial(poly) for poly in polys]
    common_indeterminants = sorted({
        indeterminant
        for poly in polys
        if not poly.isconstant()
        for indeterminant in poly.names
    })
    if not common_indeterminants:
        return polys

    for idx, poly in enumerate(polys):
        indices = numpy.array([
            common_indeterminants.index(indeterminant)
            for indeterminant in poly.names
            if indeterminant in common_indeterminants
        ])
        exponents = numpy.zeros(
            (len(poly.keys), len(common_indeterminants)), dtype=int)
        if indices.size:
            exponents[:, indices] = poly.exponents
        polys[idx] = poly.from_attributes(
            exponents=exponents,
            coefficients=poly.coefficients,
            indeterminants=common_indeterminants,
            clean=False,
        )

    return tuple(polys)


def align_exponents(*polys):
    """
    Align polynomials such that the exponents are the same.

    Aligning exponents assumes that the indeterminants is also aligned.

    Args:
        polys (numpoly.ndpoly):
            Polynomial to make adjustment to.

    Returns:
        (Tuple[numpoly.ndpoly, ...]):
            Same as ``polys``, but internal adjustments made to make them
            compatible for further operations.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> poly1 = x*y
        >>> poly2 = numpoly.polynomial([x**5, y**3-1])
        >>> print(poly1.exponents)
        [[1 1]]
        >>> print(poly2.exponents)
        [[0 0]
         [0 3]
         [5 0]]
        >>> poly1, poly2 = numpoly.align_exponents(poly1, poly2)
        >>> print(poly1)
        x*y
        >>> print(poly2)
        [x**5 -1+y**3]
        >>> print(poly1.exponents)
        [[0 0]
         [0 3]
         [1 1]
         [5 0]]
        >>> print(poly2.exponents)
        [[0 0]
         [0 3]
         [1 1]
         [5 0]]

    """
    polys = [numpoly.aspolynomial(poly) for poly in polys]
    if not all(
            polys[0].names == poly.names
            for poly in polys
    ):
        polys = list(align_indeterminants(*polys))

    global_exponents = sorted({
        tuple(exponent) for poly in polys for exponent in poly.exponents})
    for idx, poly in enumerate(polys):
        lookup = {
            tuple(exponent): coefficient
            for exponent, coefficient in zip(
                poly.exponents, poly.coefficients)
        }
        zeros = numpy.zeros(poly.shape, dtype=poly.dtype)
        coefficients = [lookup.get(exponent, zeros)
                        for exponent in global_exponents]
        polys[idx] = poly.from_attributes(
            exponents=global_exponents,
            coefficients=coefficients,
            indeterminants=poly.names,
            clean=False,
        )
    return tuple(polys)


def align_dtype(*polys):
    """
    Align polynomial by shape.

    Args:
        polys (numpoly.ndpoly):
            Polynomial to make adjustment to.

    Returns:
        (Tuple[numpoly.ndpoly, ...]):
            Same as ``polys``, but internal adjustments made to make them
            compatible for further operations.

    Examples:
        >>> x = numpoly.symbols("x")
        >>> x.dtype.name
        'int64'
        >>> poly, _ = numpoly.align_dtype(x, 4.5)
        >>> poly.dtype.name
        'float64'

    """
    polys = [numpoly.aspolynomial(poly) for poly in polys]
    dtype = numpy.sum([numpy.array(True, dtype=poly.dtype) for poly in polys]).dtype
    polys = [numpoly.ndpoly.from_attributes(
        exponents=poly.exponents,
        coefficients=poly.coefficients,
        indeterminants=poly.indeterminants,
        dtype=dtype,
        clean=False,
    ) for poly in polys]
    return tuple(polys)
